//
// Copyright 2021 The ANGLE Project Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.
//
// cl_types.h: Defines common types for the OpenCL pass-through back end.

#ifndef LIBANGLE_RENDERER_CL_CL_TYPES_H_
#define LIBANGLE_RENDERER_CL_CL_TYPES_H_

#include "libANGLE/renderer/CLtypes.h"

namespace rx
{

class CLCommandQueueCL;
class CLContextCL;
class CLDeviceCL;
class CLEventCL;
class CLKernelCL;
class CLMemoryCL;
class CLPlatformCL;
class CLProgramCL;
class CLSamplerCL;

}  // namespace rx

#endif  // LIBANGLE_RENDERER_CL_CL_TYPES_H_
